import numpy as np

with open("g1A50A.dat", "r") as file:
    data = file.readlines()

for i in range(len(data)):
    # split line at ':' and return only the second part
    data[i] = data[i].split(":")[1]
    if i > 2:
        # split line at every occurence of '->'
        data[i] = data[i].split("->")

# '.strip()' removes redundant white space around the string
entry_node = data[0].strip()
exit_node = data[1].strip()
default_out = data[2].strip()

set_of_input_chars = set([])
set_of_output_chars = set([])
set_of_nodes = set([])
transitions = dict([])

# let's process transitions;
# different inputs, outputs and nodes as added to sets
# dictionary contains transitions
for transition in data[3:]:
    current_node = transition[0].strip()
    next_node = transition[2].strip()
    input_char, output_char = transition[1].strip().split('/')

    if current_node not in set_of_nodes:
        set_of_nodes.add(current_node)
    if next_node not in set_of_nodes:
        set_of_nodes.add(next_node)
    if input_char not in set_of_input_chars:
        set_of_input_chars.add(input_char)
    if output_char not in set_of_output_chars:
        set_of_output_chars.add(output_char)

    transitions[(current_node, input_char)] = (output_char, next_node)

print("Total number of nodes is ", len(set_of_nodes))
print("The size of the input alphabet is ", len(set_of_input_chars))
print("The size of the output alphabet is ", len(set_of_output_chars), "\n")


# I rewrote sets above into lists so that you can easily manipulate them
nodes = list(set_of_nodes)
inputs = list(set_of_input_chars)
outputs = list(set_of_output_chars)
# dictionaries to recover indices of the locations of items in the above lists
node2index = dict(zip(nodes, range(len(nodes))))
input2index = dict(zip(inputs, range(len(inputs))))
output2index = dict(zip(outputs, range(len(outputs))))

# creation of the main table of shape: 'current_node X input X 2', where third dimension
# contains output and next_node, hence 2 data
main_table = np.zeros((len(set_of_nodes), len(set_of_input_chars), 2)).astype(int)
for node_index in range(len(nodes)):
    for input_index in range(len(inputs)):
        tmp = (nodes[node_index], inputs[input_index])
        if tmp in transitions:
            main_table[node_index][input_index][0] = output2index[transitions[tmp][0]]
            main_table[node_index][input_index][1] = node2index[transitions[tmp][1]]
        else:
            main_table[node_index][input_index][0] = output2index[default_out]
            main_table[node_index][input_index][1] = node_index

node_sequences = [[] for i in nodes]
node_sequences2 = [[] for i in nodes]

# unresolved is a list containing nodes that have not been resolved yet - meaning that there
# needs to be created another table in the next iteration. For more info refer to the description
# of the algorithm from lecture (slides)
unresolved = [(list(range(len(nodes))), [(i, i, i) for i in range(len(nodes))])]
while len(unresolved) > 0:
    curr_indices, curr_states = unresolved.pop()
    curr_outputs = []
    currently_visited_states = []
    for index in range(len(curr_indices)):
        # temporary variables, named this way for my lack of creativity
        tmp1 = []
        tmp2 = []
        for input_index in range(len(inputs)):
            tmp1.append(main_table[curr_states[index][input_index]][input_index][0])
            tmp2.append(main_table[curr_states[index][input_index]][input_index][1])
        curr_outputs.append(tuple(tmp1))
        currently_visited_states.append(tuple(tmp2))
    # groups will be a dictionary returning id of the group/table that individual nodes are currently in
    groups = list(set(curr_outputs))
    groups = dict(zip(groups, range(len(groups))))
    for index in range(len(curr_indices)):
        node_sequences[curr_indices[index]].append(groups[curr_outputs[index]])
        node_sequences2[curr_indices[index]].append(curr_outputs[index])
    for index in range(len(groups)):
        tmp1 = []
        tmp2 = []
        for index2 in range(len(curr_outputs)):
            if groups[curr_outputs[index2]] == index:
                tmp1.append(curr_indices[index2])
                tmp2.append(currently_visited_states[index2])
        # if the node is already alone in its group, then it's considered resolved
        if len(tmp1) > 1:
            unresolved.append((tmp1, tmp2))

zipped = list(zip(nodes, node_sequences, node_sequences2))
strings = []
for i in zipped:
    strings.append(("-".join([str(j) for j in i[1]]) + " <-> " + i[0], i[2]))

strings = sorted(strings)

print("Key to the table below:")
print("GROUP <-> NODE, [(a, b, b), ...]")
print("GROUP: the sequence of tables that were necessary to distinguish nodes in the same group")
print("NODE: name of the node")
print("[(a, b, c), ...]: is a sequence of outputs. In this case, we have a system with 3 inputs and"
      " 2 different outputs. The outputs are 'a' and 'b'.")
print("Inputs may be ", " ".join([k for k in inputs]), ".")
print("THEIR ORDER IS IMPORTANT!!!")
print("So, if you input a symbol " + str(inputs[0]) + " while in NODE, you'll get 'a' as an output.\n")

print("YOUR INPUTS ARE IN ORDER: ", " ".join([k for k in inputs]), ".")
for i in strings:
    print(i)

print(zipped[0])
